package Plugins::SmartMix::API;

use strict;
use JSON::XS::VersionOneAndTwo;
use URI::Escape qw(uri_escape uri_escape_utf8);
use Digest::MD5 qw(md5_hex);
use HTTP::Request;
use URI;

use vars qw($VERSION);

use Slim::Utils::Cache;
use Slim::Utils::Log;
use Slim::Utils::Misc;
use Slim::Utils::Prefs;

use constant BASE_URL => 'http://developer.echonest.com/api/v4/';
use constant MAX_HISTORY => 200;			# keep track of the 200 most recently played songs to pre-populate playlist history and prevent repetition

my $cache = Slim::Utils::Cache->new();
my $log   = logger('plugin.smartmix');
my $prefs = preferences('plugin.smartmix');
my $aid = '';
my $last_ratelimit = {};

sub init {
	return if $aid;

	(undef, $aid, $VERSION) = @_;

	$aid ||= $cache->get('sm_aid');
	$cache->set('sm_aid', $aid, 'never') if $aid;
}

sub catalogCreate {
	my ($class, $name, $type, $cb) = @_;
	
	$type ||= 'song';
	
	my $result = _call('catalog/create', {
		_method => 'POST',
		name    => "$name:$type",
		type    => $type,
	}, $cb) || {};
	
	return if $cb;
	
	return $result->{response}->{id} if $result->{response} && $result->{response}->{id};
	
	$log->warn('Catalog not created?');
	
	# desperate attempt to grab the ID from the error message...
	if ( $result->{response} && $result->{response}->{status} && $result->{response}->{status}->{code} == 5 && (my $message = $result->{response}->{status}->{message}) ) {
		main::INFOLOG && $log->is_info && $log->info($message);
		my ($id) = $message =~ /: ([a-z\d]+)$/i;
		$log->warn('Grabbing Catalog id from error message: ' . ($id || 'unknown'));
		return $id;
	}
}

sub catalogDelete {
	my ($class, $id, $cb) = @_;
	
	return unless $id;
	
	_call('catalog/delete', {
		_method => 'POST',
		id      => $id,
	}, $cb);
}

sub catalogUpdate {
	my ($class, $id, $data, $cb) = @_;
	
	return unless $id && $data;
	
	my $params = {
		_method => 'POST',
		id      => $id,
		data    => to_json($data),
	};
	
	return _call('catalog/update', $params, $cb);
}

sub catalogProfile {
	my ($class, $name, $type, $cb) = @_;
	
	$type ||= 'song';
	
	return _call('catalog/profile', {
		name => "$name:$type",
	}, $cb);
}

sub catalogRead {
	my ($class, $id, $start, $count, $cb) = @_;
	
	return _call('catalog/read', {
		id => $id,
		start => $start || 0, 
		results => defined $count ? $count : 100,
	}, $cb);
}

#sub playlistBasic {
#	my ($class, $id, $type, $catalog, $cb) = @_;
#	
#	my $args = {
#		type    => "$type-radio",
#		"${type}_id" => $id,
#		limit   => 'true',
#		bucket  => "id:$catalog",
#		results => 100,
#	};
#	
#	return _call('playlist/basic', $args, $cb);
#}

sub playlistDynamicCreate {
	my ($class, $args, $cb) = @_;
	
	return $class->_playlistDynamic('create', $args, sub {
		my $result = $_[0];

		# pre-populate mix with tracks we've been playing before
		if ( $result->{response} && (my $sid = $result->{response}->{session_id}) && (my $seen = $cache->get('sm_played_songs')) ) {
			$class->playlistDynamicFeedback({
				session_id => $sid,
				play_song  => $seen,
			});
		}

		$cb->(@_) if $cb;
	});
}

sub playlistDynamicRestart {
	my ($class, $args, $cb) = @_;
	
	return $class->_playlistDynamic('restart', $args, $cb);
}

sub playlistDynamicSteer {
	my ($class, $args, $cb) = @_;
	
	return $class->_playlistDynamic('steer', $args, $cb);
}

sub playlistDynamicFeedback {
	my ($class, $args, $cb) = @_;
	
	if (!$args->{session_id}) {
		$cb->() if $cb;
		return;
	}
	
	return $class->_playlistDynamic('feedback', $args, $cb);
}

sub _playlistDynamic {
	my ($class, $method, $args, $cb) = @_;
	
	if ($method =~ /(?:create|restart)/) {
		$args->{bucket} = [$args->{bucket}] if ref $args->{bucket} ne 'ARRAY' && $args->{bucket};
		$args->{bucket} ||= ['id:' . $prefs->get('songCatalogId')];
		$args->{limit}  ||= 'true' if scalar @{$args->{bucket}} == 1 && $args->{bucket}->[0] =~ /^id:CA/;
	}
	
	if ($args->{artist}) {
		$args->{artist} = [ $args->{artist} ] if ref $args->{artist} ne 'ARRAY';
		map { utf8::decode($_) } @{ $args->{artist} };
	}
	
	return _call("playlist/dynamic/$method", $args, sub {
		$cb->(@_) if $cb;
		$class->setTimestamp();
	});	
}

sub playlistDynamicNext {
	my ($class, $sid, $count, $cb) = @_;
	
	$count ||= 5;
	$count = 5 if $count > 5;
	
	my $args = {
		session_id => $sid,
		results    => $count,
	};
	
	return _call('playlist/dynamic/next', $args, sub {
		my ($result, $response) = @_;
		
		# cache result for later mapping of our ID to TEN's IDs
		if ( $result->{response} && (my $songs = $result->{response}->{songs}) ) {
			my $seen = $cache->get('sm_played_songs') || [];
			my %seen = map { $_ => 1 } @$seen;
			
			# Skip songs we've got in our play history. TEN should filter these out already, but it doesn't always work
			$songs = [ grep {
				$_->{id} && !$seen{$_->{id}}
			} @$songs ] || $result->{response}->{songs};
			
			foreach my $song ( @$songs ) {
				my $sid = $song->{id} || next;
				
				next if $seen{$sid};
				
				# keep track of songs we've played. We'll pre-populate new mixes with these to prevent repetition
				unshift @$seen, $sid;
				$seen{$sid} = 1;

				foreach ( @{$song->{foreign_ids}}, @{$song->{tracks}} ) {
					my $id = ( $_->{foreign_id} =~ /:song:(.*)/ ) ? $1 : $_->{foreign_id};
					$cache->set("sm_$id", $sid, 86400 * 30) if $id;
				}
			}

			$cache->set('sm_played_songs', [ splice @$seen, 0, MAX_HISTORY ], 'never');
	
			$result->{response}->{songs} = $songs;
		}

		# tell the callback when we failed to refresh the mix
		if ( $response && $response->code !~ /2\d\d/ ) {
			$result->{_error} = $response->error || $response->code;
		}
		
		$cb->($result);
		$class->setTimestamp();
	});
}

#sub playlistDynamicDelete {
#	my ($class, $id, $cb) = @_;
#
#	return unless $id;
#
#	return _call('playlist/dynamic/delete', {
#		session_id => $id
#	}, $cb);
#}

sub setTimestamp {
	my ($class) = @_;
	
	# set a timestamp of last activity once a day
	$class->catalogUpdate($prefs->get('songCatalogId'), [{
		action => 'update',
		catalog_keyvalues => {
			# TEN sometimes complains about invalid JSON - make sure we treat time as a string?
			timestamp => time . '',
			version   => $VERSION
		}
	}], sub {
		main::INFOLOG && $log->info('set "lastused" timestamp');
		$prefs->set('lastused', time);
	}) if ($prefs->get('lastused') || 0) + 86400 < time;
}

# XXX - we could do a lookup on the.echonest.com - using cache only for now
sub getSongId {
	my ( $class, $sid, $cb ) = @_;
	
	my $id = $cache->get("sm_$sid");
	
	return $cb ? $cb->( $id ) : $id;
}

sub last_ratelimit {
	my $t = time;
	
	if (!$last_ratelimit->{ttl} || $t > $last_ratelimit->{ttl}) {
		return {};
	}
	
	return {
		ttl       => $last_ratelimit->{ttl} - $t,
		limit     => $last_ratelimit->{'x-ratelimit-limit'},
		remaining => $last_ratelimit->{'x-ratelimit-remaining'},
		used      => $last_ratelimit->{'x-ratelimit-used'},
	};
}


sub _call {
	my ( $method, $args, $cb ) = @_;
	
	my $isPOST = (delete $args->{_method} || '') eq 'POST';
	
	my @query;
	while (my ($k, $v) = each %$args) {
		next if $k =~ /^_/;		# ignore keys starting with an underscore
		
		if (ref $v eq 'ARRAY') {
			foreach (@$v) {
				push @query, $k . '=' . uri_escape_utf8($_);
			}
		}
		else {
			push @query, $k . '=' . uri_escape_utf8($v);
		}
	}
	push @query, 'api_key=' . aid();

	my $params = join('&', @query);
	my $url = BASE_URL . $method;

	if (main::INFOLOG && $log->is_info) {
		$log->info(_debug( ($cb ? 'Async' : 'Sync') . " API call: " . ($isPOST ? 'POST' : 'GET') . " $url" . (($isPOST || !$log->is_debug) ? '' : "?$params") ));
		$log->info(_debug(Data::Dump::dump(@query))) if $isPOST;
	}
	
	# async call
	if ($cb) {
		my $cb2 = sub {
			my $response = shift;
			
			main::INFOLOG && $log->is_info && $response->code !~ /2\d\d/ && $log->info(_debug(Data::Dump::dump($response, @_)));
			my $result = _getResponseBody($response, $args);
				
			$cb->($result, $response) if ref $cb eq 'CODE';
		};
		
		my $http = Slim::Networking::SimpleAsyncHTTP->new( $cb2, $cb2, { timeout => 15 } );

		if ($isPOST) {
			$http->post($url, 'Content-Type' => 'application/x-www-form-urlencoded', $params);
		}
		else {
			$http->get($url . '?' . $params);
		}
	}
	# sync call used in scanner
	else {
		my $request;
		
		if ($isPOST) {
			$request = HTTP::Request->new( POST => $url );
			$request->content_type( 'application/x-www-form-urlencoded' );
			$request->content( $params );
		}
		else {
			$request = HTTP::Request->new( GET => $url . '?' . $params );
		}
	
		require LWP::UserAgent;
		my $ua = LWP::UserAgent->new(
			agent   => Slim::Utils::Misc::userAgentString(),
			timeout => 30,
		);
		
		my $response = $ua->request( $request );
		
		main::INFOLOG && $log->is_info && $response->code !~ /2\d\d/ && $log->info(_debug(Data::Dump::dump($response)));
		
		return _getResponseBody($response, $args);
	}
}

sub _getResponseBody {
	my ($response, $args) = @_;

	my $headers = {};
	
	if ($response && (my $h = $response->headers)) {
		while (my ($k, $v) = each %$h) {
			if ($k =~ /x-ratelimit/) {
				main::INFOLOG && $log->is_info && $log->info("$k: $v");
				$headers->{$k} = $v;
			}
		}
	}
	elsif (main::DEBUGLOG && $log->is_debug) {
		warn Data::Dump::dump($response);
	}
	
	my $t = time;
	$last_ratelimit = $headers;
	$last_ratelimit->{ttl} = $t + (60 - ($t % 60)) + 5;	# how long will the limits be valid (+5s slack)
	
	return {
		_smartmix_args => $args,
		_ratelimit     => $headers, 
	} unless $response->content;
	
	my $result = eval { from_json( $response->content ) };
	
	if ($@) {
		$log->error(Data::Dump::dump($response)) unless main::DEBUGLOG && $log->is_debug;
		$log->error($@);
	}
	main::DEBUGLOG && $log->is_debug && warn Data::Dump::dump($result);

	$result ||= {};
	$result->{_smartmix_args} = $args;
	$result->{_ratelimit}     = $headers;
	
	return $result;
}

sub _debug {
	my $msg = shift;
	$msg =~ s/$aid/\*/gi if $aid;
	return $msg;
}

sub aid {
	return ($aid =~ s/-//g) ? ($aid = uri_escape(pack('H*', $aid))) : $aid; 
}

1;