package Plugins::SmartMix::Importer;

use strict;
use File::Slurp;
use File::Spec::Functions;
use Scalar::Util qw(blessed);

use Slim::Music::Import;

use Slim::Utils::Log;
use Slim::Utils::Misc;
use Slim::Utils::Prefs;

use Plugins::SmartMix::API;

use constant MAX_UPLOAD      => 2500;
use constant MAX_UPLOAD_SAFE => 500;

use constant THROTTLING => 10;

my $prefs = preferences('plugin.smartmix');
my $log = logger('plugin.smartmix');
my $missing_log;

my $cleanup = -1;
my $newTracks = [];
my $wipe;
my $ignoreGenres;

sub initPlugin {
	my $class = shift;

	Plugins::SmartMix::API->init();

	if (!main::SCANNER) {

		require Slim::Utils::Scanner::API;
	
		Slim::Utils::Scanner::API->onNewTrack( { cb => \&onNewChangedTrack } );
		Slim::Utils::Scanner::API->onChangedTrack( { cb => \&onNewChangedTrack } );

		$prefs->setChange( sub {
			$ignoreGenres = undef;
		}, 'ignoreGenres');
	
		# Run async and throttled catalog cleanup when a scan is done.
		# This cannot be done in the scanner process, as it would easily hit the rate limit on the.echonest.com
		Slim::Control::Request::subscribe(\&stopCleanupCatalog, [['wipecache', 'rescan']] );
	}

	Slim::Music::Import->addImporter($class, {
		'type'         => 'post',
		'weight'       => 80, # make sure we run after iTunes/MusicIP, or it will miss imported files
		'reset'        => sub { $wipe = 1 }, 
		'use'          => 1,
	});
	
	$missing_log = catfile(Slim::Utils::OSDetect::dirsFor('log'), 'smartmix-missing.log');
	
	return 1;
}

sub onNewChangedTrack {
	push @$newTracks, $_[0] unless $wipe;
};

sub startScan {
	my $class = shift;

	my $catalog = $prefs->get('songCatalogId');
	
	# read flag whether we want to export a full data dump
	$wipe ||= $prefs->get('wipe');
	
	my $count;
	
	if ($catalog) {
		$count = $class->runExport($catalog);
	}

	$class->doneScanning;

	return $count;
} 

sub runExport {
	my ($class, $catalog) = @_;

	main::INFOLOG && $log->info("Start Taste Profile export");

	# always return at least one genre - even if it's an invalid one
	my $ignoreGenresIDs = join(', ', @{$class->_ignoredGenres()}, 0);

	my $dbh = Slim::Schema->dbh;
	
	my $artist_sth  = $dbh->prepare_cached( qq( SELECT name FROM contributors WHERE id=? ) );
	
	my ($persist_sth_urlmd5, $persist_sth_mb_id);
	if (main::STATISTICS) {
		$persist_sth_urlmd5 = $dbh->prepare_cached( qq( SELECT playcount, rating FROM tracks_persistent WHERE urlmd5=? ) );
		$persist_sth_mb_id  = $dbh->prepare_cached( qq( SELECT playcount, rating FROM tracks_persistent WHERE musicbrainz_id=? ) );
	}

	my $sql = "SELECT id, title, primary_artist, urlmd5, musicbrainz_id FROM tracks ";

	if ( my $library = $prefs->get('useLibrary') ) {
		eval { require Slim::Music::VirtualLibraries; };
		
		if ($@) {
			$log->warn('Virtual Library support is not available.');
		}
		else {
			$sql .= sprintf('JOIN library_track ON library_track.track = id AND library_track.library = "%s" ', $library);
		}
	}
	
	$sql .= "WHERE audio = 1 AND remote != 1 AND id NOT IN ( SELECT track FROM genre_track WHERE genre IN ($ignoreGenresIDs) GROUP BY track )"; 
	
	if (!$wipe) {
		if ( !main::SCANNER ) {
			$sql .= ' AND id IN (' . join(',', @$newTracks, 0) . ') ';
		}
		# in SCANNER mode we don't have the new/changed callbacks to count the new tracks
		elsif ( my $lastScanTime = Slim::Music::Import->lastScanTime ) {
			$sql .= " AND (timestamp > $lastScanTime OR added_time > $lastScanTime OR updated_time > $lastScanTime)";
		}
	}
	
	my ($count) = $dbh->selectrow_array( qq{
		SELECT COUNT(*) FROM ( $sql ) AS t
	}) || 0;

	my $sth = $dbh->prepare_cached($sql);
	$sth->execute();

	main::INFOLOG && $log->info("Got $count song(s) to export.");

	my $progress = Slim::Utils::Progress->new({ 
		type  => 'importer', 
		name  => 'plugin_smartmix_export', 
		total => $count, 
		bar   => 1
	}) if $count;

	my $i = 0;
	my @data;
	my $startTime = time;
	
	my $maxUpload = _maxUpload();
		
	while ( my $track = $sth->fetchrow_hashref ) {
		my $title = $track->{title};
		
		if ( $track->{primary_artist} && $title ) {
			my $item = {
				item_id   => $track->{urlmd5},
				song_name => _cleanupTitle($title),
			};

			$artist_sth->execute($track->{primary_artist});
			($item->{artist_name}) = $artist_sth->fetchrow_array;
			$item->{artist_name} = _cleanupTitle($item->{artist_name}) if $item->{artist_name};

			if ( main::STATISTICS ) {
				my ($playcount, $rating);
				if ( $track->{musicbrainz_id} ) {
					$persist_sth_mb_id->execute($track->{musicbrainz_id});
					($playcount, $rating) = $persist_sth_mb_id->fetchrow_array;
				}
				else {
					$persist_sth_urlmd5->execute($track->{urlmd5});
					($playcount, $rating) = $persist_sth_urlmd5->fetchrow_array;
				}
	
				$item->{play_count} = $playcount*1 if $playcount;
				
				# TEN ratings are 0-10, with default being 5. <4 is negative rating
				# LMS ratings are 0-100, all positive - we need to map these values
				if (defined $rating && $rating <= 100 && $rating >= 0) {
					$rating = int(($rating + 10) / 20);
					
					# we don't handle negative ratings yet
					$item->{rating} = $rating + 5;
				}
			}
			
			push @data, {
				item => $item
			};
		}

		$i++;
		
		if ( scalar @data && scalar @data % $maxUpload == 0 ) {
			my $result = Plugins::SmartMix::API->catalogUpdate($catalog, \@data);

			@data = ();

			if ( $result && ref $result eq 'HASH' && $result->{response} && $result->{response}->{status} && $result->{response}->{status}->{code} == 0 ) {
				# we're going to throttle the upload, as it could easily kill mixing due to the rate limit
				my $interval = THROTTLING - (time - $startTime);
				sleep($interval) if ($interval > 0);
				$startTime = time;
			}
			# if we didn't get a success message, cancel
			else {
				$log->error('Failed to update the taste profile:' . Data::Dump::dump($result));
				last;
			}
		}

		$progress->update($title);
		main::SCANNER && $i % 100 == 0 && Slim::Schema->forceCommit;
	}
	
	if (scalar @data) {
		Plugins::SmartMix::API->catalogUpdate($catalog, \@data);
	}

	$persist_sth_urlmd5->finish;
	$persist_sth_mb_id->finish;
	$artist_sth->finish;
	$sth->finish;

	$progress->final($count) if $count;

    return $i;
}

# try to remove additions like "feat. artist X" from artist/track names
# they often lead to tracks not being recognized
sub _cleanupTitle {
	my $item = shift;

	$item =~ s/(.+)\[.*?\]/$1/g;
	$item =~ s/(.+)\(.*(\bfeat|remix|\brmx\b|mix\b|\bwith\s|bonus ?track|extended|original|version|demo|live|instrumental|lp version).*?\)/$1/ig;
	$item =~ s/(.+)\(\d{4}\)/$1/g;

	utf8::decode($item);

	return $item;
}

sub _maxUpload {
	return $prefs->get('failsafeUpload') ? MAX_UPLOAD_SAFE : MAX_UPLOAD;
}

sub _ignoredGenres {
	return $ignoreGenres if $ignoreGenres;
	
	$ignoreGenres = [ map {
		map {
			$_->id
		} Slim::Schema->rs('Genre')->search({
			name => {
				like => '%' . $_ . '%'
			}
		})->all;
	} split(/\s*,\s*/, $prefs->get('ignoreGenres')) ];

	return $ignoreGenres;
}

sub isIgnoredGenre {
	my $track = shift;
	
	return unless blessed($track) && $track->genre;
	
	$ignoreGenres ||= { map { $_ => 1 } @{ _ignoredGenres() } };
	
	return unless keys %$ignoreGenres;
	return (grep { $ignoreGenres->{$_->id} } $track->genres) ? 1 : 0;
}

sub doneScanning {
	my $class = shift;

	$newTracks = [];
	main::INFOLOG && $log->info("Finished exporting to Taste Profile");

	Slim::Music::Import->endImporter($class);
}

# cleanup is no longer run from the importer, but the main code _after_ the import
my @deleted;
sub cleanupCatalog {
	# don't restart if we're already running
	return if isCleanupRunning();

	write_file( $missing_log, '# ' . POSIX::strftime('%d-%b-%Y %H:%M',localtime) ." - TEN Profile Cleanup started \n" );
	
	main::INFOLOG && $log->is_info && $log->info("Starting async and throttled catalog cleanup and analysis (max. " . int(60/THROTTLING) . " requests per minute)");

	@deleted = ();
	$cleanup = 0;

	Slim::Utils::Timers::killTimers(undef, \&_readCatalog);
	Slim::Utils::Timers::killTimers(undef, \&_purgeCatalog);

	Slim::Utils::Timers::setTimer(undef, time + THROTTLING + rand(2), \&_readCatalog);
}

sub stopCleanupCatalog {
	# set the number of items higher than number to process if cleanup is running
	$cleanup = 999_999 if isCleanupRunning(); 
}

sub isCleanupRunning {
	main::DEBUGLOG && $cleanup > -1 && $log->debug("Cleanup is running");
	return $cleanup > -1;
}

sub _readCatalog {
	my $catalog = $prefs->get('songCatalogId') || return;
	
	Plugins::SmartMix::API->catalogRead($catalog, $cleanup, undef, sub {
		my $response = shift;
		
		$response->{response} && (my $catalog = $response->{response}->{catalog} || {});
		
		my $items = $catalog->{items};

		if ($items && scalar @$items) {
			my @missing;

			foreach my $item (@$items) {
				my $id;
				
				if ( $item->{request} && (my $id = $item->{request}->{item_id}) ) {
					my $track = Slim::Schema->rs('Track')->find({ urlmd5 => $id });

					if (!$track || isIgnoredGenre($track)) {
						push @deleted, {
							action => 'delete',
							item   => {
								item_id => $id . ''
							}
						}; 
					}
					elsif ( !$item->{song_id} && blessed($track) ) {
						push @missing, Slim::Utils::Misc::pathFromFileURL($track->url);
						main::INFOLOG && $log->is_info && $log->info( "Song not recognized by The Echo Nest: " . Slim::Utils::Misc::pathFromFileURL($track->url) );
						main::DEBUGLOG && $log->is_debug && $log->error(Data::Dump::dump($item));
					}

				}

				$cleanup++;
				main::idleStreams();
			}

			write_file( $missing_log, {append => 1}, map { "$_\n" } @missing ) if @missing;
		}
		else {
			warn Data::Dump::dump($response);
		}

		Slim::Utils::Timers::killTimers(undef, \&_readCatalog);
		Slim::Utils::Timers::killTimers(undef, \&_purgeCatalog);

		if ($cleanup < $catalog->{total}) {
			# add 30s in case the analysis of the tracks has not started yet
			Slim::Utils::Timers::setTimer(undef, time + THROTTLING + rand(2) + ($cleanup ? 0 : 30), \&_readCatalog);
		}
		# no more items exported - we're done
		else {
			# remove duplicate entries in the list of deleted tracks
			# the API requests will fail to delete an item which has already been deleted
			my %seen;
			@deleted = grep { !$seen{ $_->{item}->{item_id} }++ } @deleted;
			
			Slim::Utils::Timers::setTimer(undef, time + THROTTLING + rand(2), \&_purgeCatalog);
		}		
	});
}

sub _purgeCatalog {
	my $catalog = $prefs->get('songCatalogId') || return;
	
	if ( $catalog && (my @items = splice(@deleted, 0, _maxUpload())) ) {
		Plugins::SmartMix::API->catalogUpdate($catalog, \@items, sub {
			Slim::Utils::Timers::killTimers(undef, \&_purgeCatalog);
			Slim::Utils::Timers::setTimer(undef, time + THROTTLING + rand(2), \&_purgeCatalog);
		});
	}
	else {
		main::INFOLOG && $log->is_info && $log->info("Catalog cleanup has finished");
		write_file( $missing_log, {append => 1}, '# ' . POSIX::strftime('%d-%b-%Y %H:%M',localtime) ." - TEN Profile Cleanup finished \n" );
		@deleted = ();
		$cleanup = -1;
	}
}

1;
