package Plugins::SmartMix::Services;

use strict;

use Slim::Utils::Cache;
use Slim::Utils::Log;
use Slim::Utils::Prefs;

my $log   = logger('plugin.smartmix');
my $prefs = preferences('plugin.smartmix');
my $cache;

my %services = (
	Spotify  => 'Plugins::SmartMix::Services::Spotify',
);

if (!$Plugins::SmartMix::Plugin::NOMYSB) {
	$services{Deezer}   = 'Plugins::SmartMix::Services::Deezer';
	$services{Rhapsody} = 'Plugins::SmartMix::Services::Rhapsody';
	$services{WiMP}     = 'Plugins::SmartMix::Services::WiMP';
}

my $activeServices;
my $trackValidators;

sub init {
	$activeServices = undef;
}

sub registerHandler {
	my ($class, $handlerClass, $name) = @_;
	
	return unless $name;

	$services{$name} = $handlerClass;
}

sub getServices {
	my ($class, $client) = @_;
	
	if (!defined $activeServices) {
		foreach (values %services) {
			my $id = $_->getId($client);
			
			$activeServices->{$id} = $_ if $id;
		}
		
		# push random service on the list if we're using an unsupported music service
		if ( (scalar keys %$activeServices) && (scalar grep { $_ !~ /^id:/ } keys %$activeServices) && !(scalar grep { $_ =~ /^id:/ } keys %$activeServices) ) {
			$activeServices->{'id:spotify'}++;
		}
		
		main::INFOLOG && $log->is_info && $log->info("List of available services: " . Data::Dump::dump($activeServices));
	}
		
	main::DEBUGLOG && $log->is_debug && $log->debug("List of available services: " . Data::Dump::dump($activeServices));

	return [ (grep /^id:/, keys %$activeServices), 'tracks' ] if $activeServices && keys %$activeServices;

	$activeServices = undef;
}

sub getAllServices {
	return [ 
		sort { 
			lc($a) cmp lc($b) 
		} grep {
			( ($services{$_} =~ /^Plugins::/ && eval('$' . $services{$_} . '::VERSION') ) 
				|| Slim::Utils::PluginManager->isEnabled("Plugins::${_}::Plugin") 
				|| Slim::Utils::PluginManager->isEnabled("Slim::Plugin::${_}::Plugin") 
				|| Slim::Utils::PluginManager->isEnabled("Slim::Plugin::${_}Logi::Plugin") 
				|| Slim::Utils::PluginManager->isEnabled("Slim::Plugin::${_}Direct::Plugin") 
			) ? $_ : undef;
		} keys %services
	];
}

sub getTrackUrl {
	my ($class, $id, $client, $alternativeTracks) = @_;
	
	my $url;
	my $handler = [];
	
	# initialize services list if needed
	$class->getServices($client) unless $activeServices;
	
	foreach ( grep /::/, values %$activeServices ) {
		
		if ($url = $_->getUrl($id, $client, $alternativeTracks)) {
			if (ref $url eq 'CODE') {
				push @$handler, $url;
				$url = undef;
			}
			else {
				last;
			}
		}
	}
	
	return $url || $handler;
}

sub getTrackIdFromUrl {
	my ($class, $url) = @_;

	my $id;
	foreach ( @{$class->getAllServices()} ) {
		if ($id = $services{$_}->urlToId($url)) {
			last;
		}
	}
	
	$cache ||= Slim::Utils::Cache->new();
	$id ||= $cache->get('sm_songid_' . $url);

	return $id;
}

sub resolveUrl {
	my $url      = pop;
	my $protocol = pop;
	my $client   = pop;

	my ($class, $cb, $args) = @_;

	Slim::Formats::XML->getFeedAsync(
		sub {
			my $feed = shift;
			
			my $candidates;
			
			# extract all potential streams
			if ($feed && $feed->{items}) {
				push @$candidates, map {
					{
						title  => $_->{line1},
						artist => $_->{line2},
						url    => $_->{play},
					}
				} grep { 
					$_->{play} && $_->{play} =~ /^\Q$protocol\E/
				} @{$feed->{items}};
			}

			$cb->( $class->getUrlFromCandidates($candidates, $args) );
		},
		sub { $cb->() },
		{
			client => $client,
			url => Slim::Networking::SqueezeNetwork->url( sprintf($url, $args->{title}) ),
		},
	);
}

sub skipTrack {
	my ( $class, $client, $track ) = @_;
	
	my $url;
	return 1 if !( $track && blessed $track && ($url = $track->url) );
	
	main::DEBUGLOG && $log->debug("Validating track $url");
	
	if (!$trackValidators) {
		$trackValidators = [];
			
		# initialize services list if needed
		$class->getServices($client) unless $activeServices;
		
		foreach ( grep /::/, values %$activeServices ) {
			push @$trackValidators, $_ if $_->can('skipTrack');
		}
	}
	
	foreach ( @$trackValidators ) {
		# only call validator if track is of the given class - prevents expensive db lookups
		next unless $_->urlToId($url);

		return 1 if $_->skipTrack($client, $track);
	}
}

sub getUrlFromCandidates {
	my ($class, $candidates, $args) = @_;

	my $artist = lc($args->{artist});
	my $title  = lc($args->{title});

	# XXX - apply some more smarts, like eg. "Tony Levin Band" vs. "Tony Levin", "Remaster" etc.
	# do some title/artist cleanup
	$candidates = [ grep {
		$_->{title} !~ /karaoke|sound.a.like|as.made.famous/ && $_->{artist} !~ /karaoke/
	} map {
		my $title = $class->cleanupTitle( lc($_->{title}) );
		my $artist = lc($_->{artist});
		
		# artist comes from a menu with "artist - album" in it
		$artist =~ s/ - .*//;

		{
			title  => lc($title),
			artist => lc($artist),
			url    => $_->{url},
		}
	} @$candidates ];
	
	if (main::INFOLOG && $log->is_info) {
		$log->info("Trying to match criteria: " . Data::Dump::dump($args));
		$log->info( Data::Dump::dump(
			map { "$_->{title} - $_->{artist}" } 
			grep { $_->{title} =~ /^\Q$title\E/ || $title =~ /^\Q$_->{title}\E/ }
			@$candidates)
		);
	} 
	
	# we don't care about tracks which don't even match the start of the title name
	my @candidates = grep { 
		$_->{title} =~ /^\Q$title\E/
	} @$candidates;

	# find match for "artist" - "title"
	my ($url) = map { $_->{url} } grep { $_->{title} eq $title && $_->{artist} eq $artist } @candidates;
	# "artist" - "title *"
	($url) = map { $_->{url} } grep { $_->{artist} eq $artist } @candidates unless $url;
	# "artist *" - "title"
	($url) = map { $_->{url} } grep { $_->{title} eq $title && $_->{artist} =~ /^\Q$artist\E/ } @candidates unless $url;
	# "artist *" - "title *"
	($url) = map { $_->{url} } grep { $_->{artist} =~ /^\Q$artist\E/ } @candidates unless $url;
	($url) = map { $_->{url} } grep { $artist =~ /^\Q$_->{artist}\E/ } @candidates unless $url;
	
	@candidates = grep { 
		$title =~ /^\Q$_->{title}\E/ 
	} @$candidates unless $url;
	
	# "title *" - "artist *"
	($url) = map { $_->{url} } grep { $title =~ /^\Q$_->{title}\E/ && $_->{artist} =~ /^\Q$artist\E/ } @candidates unless $url;

	if (main::INFOLOG && $log->is_info) {
		if ($url) {
			$log->info("Chose $url for " . Data::Dump::dump(grep {$_->{url} eq $url} @$candidates));
		}
		else {
			$log->info("No match!");
		}
	} 
	
	$cache ||= Slim::Utils::Cache->new();
	$cache->set('sm_songid_' . $url, $args->{id}, 86400*7) if $url && $args->{id};
	
	return $url;
}

sub cleanupTitle {
	my ($class, $title) = @_;

	# remove everything between () or []... But don't for PG's eponymous first four albums :-)
	$title =~ s/[\(\[].*?[\)\]]//g;
	
	# remove stuff like "2012", "live"
	$title =~ s/\d+\/\d+//ig;
	$title =~ s/- live\b//i;
	$title =~ s/- remaster.*//i;

	# remove trailing non-word characters
	$title =~ s/[\s\W]{2,}$//;
	$title =~ s/\s*$//;
	
	return $title;
}

1;


package Plugins::SmartMix::Services::Spotify;

use JSON::XS::VersionOneAndTwo;

my $use3rdPartySpotify;
my $noSSL;

# Spotify country codes available (http://developer.echonest.com/docs/v4#project-rosetta-stone)
my %spotifyCCs = map {
	$_ => $_
} qw(AD AR AT AU BE BG BO BR CA CH CL CO CR CY CZ DE DK DO EC EE ES FI FR GB GR GT HK HN HR HU IE IS IT LI LT LU LV MC MT MX MY NI NL NO NZ PA PE PH PL PT PY RO SE SG SI SK SV TR TW US UY ZZ);

sub getId {
	my ($class, $client) = @_;

	if (!defined $noSSL) {
		# use proxied request if no SSL is available
		$noSSL = Slim::Networking::Async::HTTP::hasSSL() ? 0 : 1;
	}

	return unless $client;
	
	if (!defined $use3rdPartySpotify) {
		$use3rdPartySpotify = (Slim::Utils::PluginManager->isEnabled('Plugins::Spotify::Plugin') || Slim::Utils::PluginManager->isEnabled('Plugins::SpotifyProtocolHandler::Plugin')) ? 1 : 0;
	}
	
	return if $Plugins::SmartMix::Plugin::NOMYSB && !$use3rdPartySpotify; 

	return unless $use3rdPartySpotify || Slim::Utils::PluginManager->isEnabled('Slim::Plugin::SpotifyLogi::Plugin');
	
	return unless $use3rdPartySpotify || $client->isAppEnabled('Spotify');
	
	return if $prefs->get('disable_Spotify');

	# spotify on ip3k only with Triode's plugin
	return unless $use3rdPartySpotify || ($client->isa('Slim::Player::SqueezePlay') && $client->model ne 'squeezeplay');

	my $cc = $class->getSpotifyCountry();
	
	return 'id:spotify' . ($cc ne 'WW' ? "-$cc" : '');
} 

sub getUrl {
	my ($class, $id, $client, $alternativeTracks) = @_;

	if ( $id =~ /(spotify)(?:-[A-Z]{2})?:(track:.*)/ ) {
		return $use3rdPartySpotify ? "$1:$2" : "$1://$2";
	}

=pod
	Now that TEN give access to the Spotify country specific buckets, we no longer need this

	my $cc = $class->getSpotifyCountry('XX');
	if ( !$cc || $cc eq 'XX' ) {
		main::INFOLOG && $log->is_info && $log->info("No spotify country set, returning track: " . $id);
	
		if ( $id =~ /(spotify)(?:-[A-Z]{2})?:(track:.*)/ ) {
			return $use3rdPartySpotify ? "$1:$2" : "$1://$2";
		}
		else {
			return;
		}
	}

	$alternativeTracks ||= [];

	# extract IDs only from the track list
	@$alternativeTracks = map {
		$_->{foreign_id} =~ /spotify(?:-[A-Z]{2})?:track:(.*)/;
		$1;
	} grep {
		$_->{catalog} =~ /spotify/i
	} @$alternativeTracks;

	if ( $id =~ /spotify(?:-[A-Z]{2})?:track:(.*)/ ) {
		push @$alternativeTracks, $1;
	}

	main::INFOLOG && $log->is_info && $log->info("Storing list of alternative spotify tracks to check: " . Data::Dump::dump($alternativeTracks));
	
	return unless scalar @$alternativeTracks;
	
	$client->pluginData( trackIds => $alternativeTracks );
	
	# we can't handle the id - return a search handler instead
	return sub {
		$class->resolveUrl(@_, $client);
	};
=cut 
}

=pod
sub resolveUrl {
	my $client = pop;
	my ($class, $cb, $args) = @_;
	
	my $trackIds = join(',', grep { $_ } @{ $client->pluginData('trackIds') || [] }[0..40]);

	main::DEBUGLOG && $log->is_debug && !$trackIds && $log->debug("No spotify track IDs found?!?");
	
	if (!$trackIds) {
		$cb->();
		return;
	}

	main::DEBUGLOG && $log->is_debug && $log->debug("Going to check availability for spotify track IDs: $trackIds");

	Slim::Networking::SimpleAsyncHTTP->new(
		sub {
			my ($response) = @_;
		
			my $result = eval { from_json( $response->content ) };
			
			if ($@) {
				$log->error($@);
				$cb->();
			}
			else {
				main::DEBUGLOG && $log->is_debug && $log->debug(Data::Dump::dump($result));
			
				my $cc = $class->getSpotifyCountry('US');
				my $trackId;
				foreach my $track ( sort { $b->{popularity} <=> $a->{popularity} } @{ $result->{tracks} || [] }) {
					
					next unless $track && ref $track eq 'HASH';

					if ( $track->{available_markets} && grep /$cc/, @{$track->{available_markets}} ) {
						main::DEBUGLOG && $log->is_debug && $log->debug("found valid track: " . Data::Dump::dump($track));
						$trackId = $track->{id};
						last;
					}
				}

				$trackId = ( $use3rdPartySpotify ? "spotify:track:$trackId" : "spotify://track:$trackId" ) if $trackId;

				main::INFOLOG && !$trackId && $log->is_info && $log->info("Didn't find playable track for country $cc and track: " . Data::Dump::dump($args)); 

				$cb->($trackId);
			}
		},
		sub {
			$log->error("Something went wrong checking Spotify availability: " . Data::Dump::dump(@_));
			$cb->();
		},{
			timeout => 15
		}
	)->get( $noSSL ? sprintf('http://www.herger.net/cgi-bin/spotifytracks.cgi?ids=%s&amp;f=1', $trackIds)  : ('https://api.spotify.com/v1/tracks?ids=' . $trackIds) );
}
=cut

sub urlToId {
	my ($class, $url) = @_;
	
	if ($url =~ m|(spotify):/{0,2}(track:.*)|) {
		my $cc = $class->getSpotifyCountry();
		return "$1-$cc:$2";
	}
}

sub getSpotifyCountry {
	my ($class, $default) = @_;

	my $cc = $prefs->get('spotifyCountry');
	$cc = $spotifyCCs{$cc} if $cc;
	return $cc || $default || 'WW';
}

1;


package Plugins::SmartMix::Services::Deezer;

use Plugins::SmartMix::Mixer;

sub getId {
	my ($class, $client) = @_;

	return if $Plugins::SmartMix::Plugin::NOMYSB;
	
	return unless $client;
	return unless Slim::Utils::PluginManager->isEnabled('Slim::Plugin::Deezer::Plugin');

	return ( $client->isAppEnabled('Deezer') && !$prefs->get('disable_Deezer') ) ? 'deezer' : undef;
} 

sub getUrl {
	my ($class, $id, $client) = @_;
	
	# we can't handle the id - return a search handler instead
	return sub {
		Plugins::SmartMix::Services->resolveUrl(@_, $client, 'deezer:', '/api/deezer/v1/opml/search_tracks?q=%s');
	} if $class->getId($client); 
}

# dealt with in Plugins::SmartMix::Services->getTrackIdFromUrl
sub urlToId {}

sub skipTrack {
	my ($class, $client, $track) = @_;
	
	my ($artist, $title, $duration) = Plugins::SmartMix::Mixer::getMixablePropertiesFromTrack($client, $track);
	
	# skip samples (length = 30s)
	return $duration == 30 ? 1 : 0;
}
	
1;


package Plugins::SmartMix::Services::Rhapsody;

sub getId {
	my ($class, $client) = @_;
	
	return if $Plugins::SmartMix::Plugin::NOMYSB;
	
	return if !$client;
	
	return if !$client->isa('Slim::Player::Squeezebox2') || $client->model eq 'squeezeplay';
	
	return if $prefs->get('disable_Rhapsody');

	return unless Slim::Utils::PluginManager->isEnabled('Slim::Plugin::RhapsodyDirect::Plugin');
	
	return ($client->isAppEnabled('RhapsodyDirect') || $client->isAppEnabled('RhapsodyEU')) ? 'id:rhapsody-US' : undef;
} 

sub getUrl {
	my ($class, $url) = @_;

	return unless Slim::Utils::PluginManager->isEnabled('Slim::Plugin::RhapsodyDirect::Plugin');
	
	my $ext = Slim::Plugin::RhapsodyDirect::ProtocolHandler::getFormatForURL();
	
	$ext = 'm4a' if $ext eq 'mp4';
		
	if ($url =~ /rhapsody-..:track:(.*)/) {
		return "rhapd://$1.$ext";
	}
}

sub urlToId {
	my ($class, $url) = @_;
	
	if ($url =~ m{rhapd://(.*)\.(?:mp3|m4a)}) {
		# XXX - not 100% accurate, but we don't know where the user is from (US vs. EU);
		return "rhapsody-US:track:$2";
	}
}

1;


package Plugins::SmartMix::Services::WiMP;

use Slim::Utils::Log;
use Slim::Utils::Prefs;

use Plugins::SmartMix::Services;

my $log   = logger('plugin.smartmix');
my $prefs = preferences('plugin.smartmix');

sub getId {
	my ($class, $client) = @_;

	return if $Plugins::SmartMix::Plugin::NOMYSB;
	
	return unless $client;
	return unless Slim::Utils::PluginManager->isEnabled('Slim::Plugin::WiMP::Plugin');
	
	return ( ($client->isAppEnabled('WiMP') || $client->isAppEnabled('WiMPDK')) && !$prefs->get('disable_WiMP') ) ? 'wimp' : undef;
} 

sub getUrl {
	my ($class, $id, $client) = @_;
	
	# we can't handle the id - return a search handler instead
	return sub {
		Plugins::SmartMix::Services->resolveUrl(@_, $client, 'wimp:', '/api/wimp/v1/opml/search?q=%s');
	} if $class->getId($client); 
}

# dealt with in Plugins::SmartMix::Services->getTrackIdFromUrl
sub urlToId {}

1;